{***************************************************************************}
{ TDBFinderEdit component by Deborah Pate                                   }
{ Version 1.81, May 1999                                                    }
{                                                                           }
{ A data-aware edit control for finding records in Delphi 4.                }
{ It uses the Locate method of a dataset, so works on both indexed and      }
{ unindexed fields.                                                         }
{                                                                           }
{ N.B. You can't edit records with this control - though obviously you can  }
{ edit them with other controls once you've found them with this ...        }
{                                                                           }
{***************************************************************************}
unit DBFEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics, StdCtrls,
  DB, DbTables, DBCtrls, DBFFuncs;

type
  TDBFinderEdit = class(TCustomEdit)
  private
    fAutoClear: boolean;
    fChanging: boolean;
    fDataLink: TFieldDataLink;
    fEnterAsTab: boolean;
    fFilterMode: TFilterMode;
    fFound: boolean;
    fGroupIndex: cardinal;
    fIncrementalSearch: boolean;
    fLocateOptions: TLocateOptions;
    fMustFind: boolean;
    fOnFound: TNotifyEvent;
    fOnFoundElsewhere: TNotifyEvent;
    fOnNotFound: TNotifyEvent;
    fOnMouseExit: TNotifyEvent;
    fOnMouseEnter: TNotifyEvent;
  protected
    { Field access methods }
    function GetDataSource: TDataSource;
    procedure SetDataSource(DataSource: TDataSource);
    function GetDataField: string;
    procedure SetDataField(const Field: string);
    procedure SetOnMouseEnter(const Value: TNotifyEvent);
    procedure SetOnMouseExit(const Value: TNotifyEvent);
    function GetValue: variant;
    { Other methods }
    procedure ClearYoungerSibs; virtual;
    procedure Click; override;
    procedure CMGetDataLink(var Msg: TMessage); message CM_GETDATALINK;
    procedure CMMouseEnter (var Msg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseExit (var Msg: TMessage); message CM_MOUSELEAVE;
    procedure DoExit; override;
    procedure DoLocate; virtual;
    procedure DoMouseEnter; virtual;
    procedure DoMouseExit; virtual;
    procedure KeyPress(var Key: char); override;
    procedure KeyUp(var Key: word; Shift: TShiftState); override;
    procedure Notification(Component: TComponent; Operation: TOperation); override;
    procedure RecordChange(Sender: TObject); virtual;
    property Changing: boolean read fChanging write fChanging;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Found: boolean read fFound write fFound;
    property Value: variant read GetValue;
  published
    property AutoClear: boolean read fAutoClear write fAutoClear default False;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DataField: string read GetDataField write SetDataField;
    property EnterAsTab: boolean read fEnterAsTab write fEnterAsTab default False;
    property FilterMode: TFilterMode read fFilterMode write fFilterMode default fmIgnoreFilter;
    property GroupIndex: cardinal read fGroupIndex write fGroupIndex default 0;
    property IncrementalSearch: boolean read fIncrementalSearch write fIncrementalSearch default True;
    property LocateOptions: TLocateOptions read fLocateOptions write fLocateOptions default [loCaseInsensitive,loPartialKey];
    property MustFind: boolean read fMustFind write fMustFind default False;
    { Events }
    property OnFound: TNotifyEvent read fOnFound write fOnFound;
    property OnFoundElsewhere: TNotifyEvent read fOnFoundElsewhere write fOnFoundElsewhere;
    property OnMouseEnter: TNotifyEvent read fOnMouseEnter write SetOnMouseEnter;
    property OnMouseExit: TNotifyEvent read fOnMouseExit write SetOnMouseExit;
    property OnNotFound: TNotifyEvent read fOnNotFound write fOnNotFound;
    { Inherited properties }
    {$IFDEF VER120}
      property Anchors;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ImeMode;
      property ImeName;
      property ParentBiDiMode;
      property OnEndDock;
      property OnStartDock;
    {$ENDIF}
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    { Inherited events }
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
 end;

implementation

uses Forms;

{ Construction and destruction }

constructor TDBFinderEdit.Create(AOwner: TComponent);
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(AOwner <> nil, 'TDBFinderEdit.Create can''t accept a nil parameter in its constructor');
  {$ENDIF}
  inherited Create(AOwner);

  { Set up the link to the data }
  fDataLink := TFieldDataLink.Create;
  fDataLink.Control := Self;
  fDataLink.OnDataChange := RecordChange;

  { Initialize properties }
  fAutoClear := False;
  fChanging := False;
  fEnterAsTab := False;
  fFilterMode := fmIgnoreFilter;
  fFound := False;
  fGroupIndex := 0;
  fIncrementalSearch := True;
  fLocateOptions := [loCaseInsensitive,loPartialKey];
  fMustfind := False;
end;

destructor TDBFinderEdit.Destroy;
begin
  fDataLink.Free;
  fDataLink := nil;
  inherited Destroy;
end;

{ Property access methods }

function TDBFinderEdit.GetDataSource: TDataSource;
begin
  Result := fDataLink.DataSource;
end;

function TDBFinderEdit.GetDataField: string;
begin
  Result := fDataLink.FieldName;
end;

function TDBFinderEdit.GetValue: variant;
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert((fDataLink.Field <> nil), 'TDBFinderEdit.GetValue: Field is nil');
  {$ENDIF}
  Result := fDataLink.Field.Value;
end;

procedure TDBFinderEdit.SetDataSource(DataSource: TDataSource);
begin
  fDataLink.DataSource := DataSource;
  { Request notification if the DataSource component is removed from the form }
  if DataSource <> nil then
    DataSource.FreeNotification(Self);
end;

procedure TDBFinderEdit.SetDataField(const Field: string);
begin
  fDataLink.FieldName := Field;
end;

procedure TDBFinderEdit.SetOnMouseEnter(const Value: TNotifyEvent);
begin
  fOnMouseEnter := Value;
end;

procedure TDBFinderEdit.SetOnMouseExit(const Value: TNotifyEvent);
begin
  fOnMouseExit := Value;
end;

{****************************************************************************}
{ Other methods }

procedure TDBFinderEdit.ClearYoungerSibs;
{ If you are searching on multiple fields and set AutoClear to True, then this
  function will be called when you change the text in a TDBFinder. It finds
  other TDBFinders with the same parent and GroupIndex, and clears any that are
  lower down the TabOrder. This is useful if you don't want your search still to
  be limited to, e.g., people with first name 'John', when you've changed the
  surname you're searching for. }
var
  i: cardinal;
  SibInfo: TDBFinderInfo;
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(Assigned(Parent), 'TDBFinderEdit.ClearYoungerSibs: Control needs parent');
  {$ENDIF}
  for i := 0 to Parent.ControlCount - 1 do begin
    SibInfo := GetDBFinderInfo(Parent.Controls[i]);
    if (SibInfo.IsAFinder) then begin
      if ((SibInfo.TabOrder > TabOrder) and (SibInfo.GroupIndex = GroupIndex)) then
         ClearDBFinder(Parent.Controls[i]);
    end; { if }
  end; { for }
end;

procedure TDBFinderEdit.Click;
{ Make changing the control's text faster by making sure it's selected }
begin
  inherited Click;
  SelectAll;
end;

procedure TDBFinderEdit.CMGetDataLink(var Msg: TMessage);
{ Data-aware controls must respond to the Cm_GetDataLink message by returning
  a datalink reference }
begin
  Msg.Result := LongInt(fDataLink);
end;

procedure TDBFinderEdit.CMMouseEnter(var Msg: TMessage);
begin
  inherited;
  DoMouseEnter;
end;

procedure TDBFinderEdit.CMMouseExit(var Msg: TMessage);
begin
  inherited;
  DoMouseExit;
end;

procedure TDBFinderEdit.DoExit;
{ If we haven't been searching incrementally, now's the time to do it; if we
  have, make sure the control shows the whole of the value found. }
begin
  inherited DoExit;
  if not IncrementalSearch then begin
    if ((GroupIndex > 0) and AutoClear) then
      ClearYoungerSibs;
    DoLocate;
  end
  else if (Found and (Text <> '')) then
    Text := fDataLink.Field.DisplayText;

  { Prevent departure if record must be found }
  if ((fMustFind) and not (Found)) then begin
    SetFocus;
  end
  else begin
    Modified := False;
  end;
end;

procedure TDBFinderEdit.DoLocate;
{ Here's where we do the search }
var
  i: cardinal;
  FindFields: string;
  Temp, FindValues: variant;
  SibCount: cardinal;
  SibInfo: TDBFinderInfo;
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
     Assert((fDatalink.Datasource.Dataset.Active), 'TDBFinderEdit.DoLocate: table must be active');
  {$ENDIF}
  { Sort out the search parameters for a single field }
  if (GroupIndex = 0) then begin
    if (Text = '') then begin
      Found := False;
      Exit;
    end;

    FindFields := DataField;
    if FindFields = '' then
      raise EDBFinderError.Create(SNoDatafield);
    FindValues := Text;
  end

  { Sort out the search parameters for multiple fields }
  else begin
    SibCount := 0;
    FindValues := VarArrayCreate([0,Parent.ControlCount],varVariant);

    { Go through the Parent's controls looking for TDBFinders }
    for i := 0 to Parent.ControlCount - 1 do begin
      SibInfo := GetDBFinderInfo(Parent.Controls[i]);

      if (SibInfo.IsAFinder) then begin

        { Make sure it's in the same group, with equal or higher priority,
          and with a non-empty text property }
        if ((SibInfo.GroupIndex = GroupIndex) and
          ((not AutoClear) or (SibInfo.TabOrder <= TabOrder)) and
          (SibInfo.Text <> '')) then begin

          { If so, add its Fieldname and value to the search criteria }
          if SibCount > 0 then
             FindFields := FindFields + ';';
          FindValues[SibCount] := SibInfo.Text;
          FindFields := FindFields + SibInfo.Datafield;
          Inc(SibCount);
        end;
      end; { if }

    end; { for }

    if SibCount = 0 then begin
       Found := False;
       Exit;
    end;

    { Get around the bug in Locate which doesn't allow single value arrays.
      Note that this code has changed from the Delphi 2 version, which causes
      an exception. }
    if SibCount = 1 then begin
      Temp := FindValues[0];
      VarClear(FindValues);
      FindValues := Temp;
    end
    else VarArrayRedim(FindValues,SibCount - 1);

  end; { GroupIndex > 0 }

  { Sort out the filter mode }
  with DataSource.Dataset do begin
    case FilterMode of
      fmApplyFilter: if Filtered = False then Filtered := True;
      fmRemoveFilter: if Filtered = True then Filtered := False;
    end;
  end;

  { Set the Changing flag so the RecordChange method knows it's this constrol
    that's changed the record }
  Changing := True;

  { Do the find - at last! }
  Found := DataSource.Dataset.Locate(FindFields,FindValues,LocateOptions);

  { Allow post-search events }
  if Found then begin
    if Assigned(FOnFound) then FOnFound(Self);
  end
  else begin
    Changing := False;
    if Assigned(FOnNotFound) then FOnNotFound(Self);
  end;
end;

procedure TDBFinderEdit.DoMouseEnter;
begin
  if Assigned(fOnMouseEnter) then
    fOnMouseEnter(Self);
end;

procedure TDBFinderEdit.DoMouseExit;
begin
  if Assigned(fOnMouseExit) then
    fOnMouseExit(Self);
end;

procedure TDBFinderEdit.KeyPress(var Key: char);
{ Trap the Return and Escape keys, and make them behave as required; beep and
  do nothing if a key that isn't valid for the associated field is pressed }
begin
  {$IFOPT D+} { needed for Delphi2 compatibility }
    Assert(Owner <> nil, 'TDBFinderEdit.KeyPress: control must have an owner');
  {$ENDIF}
  if fDataLink.Field = nil then
    raise EDBFinderError.Create(SNoDatafield);
  case Key of
    { return key pressed }
    #13: begin
           Key := #0;
           if EnterAsTab then
             PostMessage((Owner as TForm).Handle,WM_NEXTDLGCTL, 0, 0)
           else SelectAll;
         end;
     { escape key pressed }
    #27: begin
           Clear;
           if ((GroupIndex > 0) and AutoClear) then
             ClearYoungerSibs;
           Found := False;
           Key := #0;
         end;
     { catch invalid keys }
    #32..#255: if not fDatalink.Field.IsValidChar(Key) then begin
                 MessageBeep(Cardinal(-1));
                 Key := #0;
               end;
  end; { case }
  inherited KeyPress(Key);
end;

procedure TDBFinderEdit.KeyUp(var Key: word; Shift: TShiftState);
{ If a valid key has been pressed and IncrementalSearch is set to True, then
  call ClearYoungerSibs if appropriate and do the find }
begin
  inherited KeyUp(Key,Shift);
  if (IncrementalSearch and (Key in [vk_Back, 32..255])) then
  begin
    if (AutoClear and (GroupIndex > 0)) then
      ClearYoungerSibs;
    DoLocate;
  end;
end;

procedure TDBFinderEdit.Notification(Component: TComponent; Operation: TOperation);
{ If the DataSource component is removed from the form, clear the
  DataSource pointer in the datalink }
begin
  inherited Notification(Component, Operation);
  if (fDataLink <> nil)
    and (Component = DataSource)
    and (Operation = opRemove) then
    DataSource := nil;
end;

procedure TDBFinderEdit.RecordChange(Sender: TObject);
{ If the record's been changed by another control, update the display }
begin
  if Changing then begin
    Changing := False;
    Exit;
  end;
  if (fDataLink.Field = nil) or (DataSource.State = dsInsert) then begin
    Text := '';
    Enabled := False;
  end
  else begin
    Enabled := not fDataLink.Editing;
    Text := fDataLink.Field.DisplayText;
    if (Showing and (Assigned(fOnFoundElsewhere))) then
      fOnFoundElsewhere(Self);
  end;
end;


end.
